(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2007-2009                                               *)
(*    CEA (Commissariat  l'nergie Atomique)                             *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

(** Bunch of values which may be initialized through command line. *)

open Extlib
open Plugin

(* ************************************************************************* *)
(** {2 General purpose options} *)
(* ************************************************************************* *)

let check_range name ~min ~max v =
  if v < min || v > max then
    Kernel.abort 
      "invalid argument for %s option, not in range %d-%d"
      name min max

(** General parameters *)
module General = struct

  include Kernel

  let () = Plugin.set_negative_option_name ""
  module Quiet =
    Bool(struct
	   let default = Cmdline.quiet
	   let option_name = "-quiet"
	   let module_name = "Quiet"
	   let descr = "sets -verbose and -debug to 0"
	 end)

  let () = set_cmdline_stage Cmdline.Exiting
  module PrintVersion =
    False
      (struct
	 let option_name = "-version"
	 let module_name = "PrintVersion"
	 let descr = "print version information"
       end)
  let () = PrintVersion.add_alias [ "-v"; "--version" ]

  let () = set_cmdline_stage Cmdline.Exiting
  module PrintShare =
    False(struct
	    let option_name = "-print-path"
	    let module_name = "PrintShare"
	    let descr = "print the Frama-C share path"
	  end)

  let () = set_cmdline_stage Cmdline.Exiting
  module PrintLib =
    False(struct
	    let option_name = "-print-libpath"
	    let module_name = "PrintLib"
	    let descr = "print the path of the Frama-C kernel library"
	  end)

  let () = set_cmdline_stage Cmdline.Exiting
  module PrintPluginPath =
    False(struct
	    let option_name = "-print-plugin-path"
	    let module_name = "PrintPluginPath"
	    let descr =
	      "print the path where the Frama-C dynamic plug-ins are installed"
	  end)

  let () = set_cmdline_stage Cmdline.Extending
  module AddPath =
    StringSet(struct
		let option_name = "-add-path"
		let module_name = "AddPath"
		let arg_name = "p1, ..., pn"
		let descr = "add paths which dynamic plugins are searched in"
	      end)
  let () = AddPath.add_set_hook (fun _ _ -> AddPath.iter Dynamic.add_path)

  let () = set_cmdline_stage Cmdline.Extending
  module LoadModule =
    StringSet(struct
		let option_name = "-load-module"
		let module_name = "LoadModule"
		let arg_name = "m1, ..., mn"
		let descr = "load the given modules dynamically"
	      end)
  let () =
    LoadModule.add_set_hook (fun _ _ -> LoadModule.iter Dynamic.load_module)

  let () = set_cmdline_stage Cmdline.Extending
  module LoadScript =
    StringSet(struct
		let option_name = "-load-script"
		let module_name = "LoadScript"
		let arg_name = "m1, ..., mn"
		let descr = "load the given ocaml scripts dynamically"
	      end)
  let () =
    LoadScript.add_set_hook (fun _ _ -> LoadScript.iter Dynamic.load_script)

  let () = do_not_journalize ()
  module UseUnicode =
    True
      (struct
	 let option_name = "-unicode"
	 let module_name = "UseUnicode"
	 let descr = "use utf8 in messages"
       end)
  let () = UseUnicode.add_set_hook (fun _ b -> Cil.print_utf8 := b)

  module Obfuscate =
    False
      (struct
	 let option_name = "-obfuscate"
	 let module_name = "Obfuscate"
	 let descr = "print an obfuscated version of files and exit"
       end)


  let () = set_negative_option_name "-unsafe-arrays"
  module SafeArrays =
    True(struct
	    let module_name = "SafeArrays"
	    let option_name = "-safe-arrays"
	    let descr = "for arrays that are fields inside structs, assume that accesses are in bounds"
	  end)

  module UnrollingLevel =
    Zero
      (struct
	 let module_name = "UnrollingLevel"
	 let option_name = "-ulevel"
	 let arg_name = "l"
	 let descr = "unroll loops n times (defaults to 0) before analyzes"
       end)

  module MainFunction =
    String(struct
	     let module_name = "MainFunction"
	     let default = "main"
	     let option_name = "-main"
	     let arg_name = "f"
	     let descr = "set to name the entry point for analysis. Use -lib-entry if this is not for a complete application. Defaults to main"
	   end)

  module LibEntry =
    False(struct
	    let module_name = "LibEntry"
	    let option_name = "-lib-entry"
	    let descr ="run analysis for an incomplete application e.g. an API call. See the -main option to set the entry point name"
	  end)

  module Machdep =
    EmptyString
      (struct
	 let module_name = "Machdep"
	 let option_name = "-machdep"
	 let arg_name = "machine"
	 let descr = "use <machine> as the current machine dependent configuration. Use -machdep help to see the list of available machines"
       end)
  let () = Project.Computation.add_dependency Cil.selfMachine Machdep.self

  module PrintCode =
    False(struct
	    let module_name = "PrintCode"
	    let option_name = "-print"
	    let descr = "pretty print original code with its comments"
	  end)

  module PrintComments =
    False(struct
	    let module_name = "PrintComments"
	    let option_name = "-keep-comments"
	    let descr = "try to keep comments in C code"
	  end)

  module CodeOutput = struct

    include EmptyString
      (struct
	 let module_name = "CodeOutput"
	 let option_name = "-ocode"
	 let arg_name = "filename"
	 let descr =
	   "when printing code, redirects the output to file <filename>"
       end)

    let state = ref (Format.std_formatter, stdout)
    let get_fmt () = fst !state

    let close old_filename =
      if old_filename <> "" then begin
	let fmt, cout = !state in
	Format.pp_print_flush fmt ();
	close_out cout
      end

    let () = at_exit (fun () -> close (get ()))

    let set_channel old_filename out_cmdline =
      close old_filename;
      let new_state =
	if out_cmdline = "" then
	  Format.std_formatter, stdout
	else
	  try
	    let out = open_out out_cmdline in
	    let fmt = Format.formatter_of_out_channel out in
	    fmt, out
	  with Sys_error s ->
	    Kernel.warning
	      "could not open file \"%s\" for code output (system error: %s).\
I will output the code on stdout instead."
	      out_cmdline s;
	    Format.std_formatter, stdout
      in
      state := new_state

    let () = add_set_hook set_channel

    let () =
      let only =
	Project.Selection.singleton self Kind.Do_Not_Select_Dependencies
      in
      Project.register_after_set_current_hook ~user_only:false
	(fun p -> (* set [state] to the right value *)
	   set (Project.on ~only p get ()))

  end

  module Time =
    EmptyString
      (struct
	 let module_name = "Time"
	 let option_name = "-time"
	 let arg_name = "filename"
	 let descr = "append user time and date to <filename> at exit"
       end)

  module SaveState =
    EmptyString
      (struct
	 let module_name = "SaveState"
	 let option_name = "-save"
	 let arg_name = "filename"
	 let descr = "save the state into file <filename> after computations"
       end)

  let () = set_cmdline_stage Cmdline.Loading
  module LoadState =
    EmptyString
      (struct
	 let module_name = "LoadState"
	 let option_name = "-load"
	 let arg_name = "filename"
	 let descr =
	   "load an initial (previously saved) state from file <filename>"
       end)

  module Overflow =
    True(struct
	  let module_name = "Overflow"
	  let option_name = "-overflow"
	  let descr = "assume that arithmetic operations overflow"
	  end)

  module UnspecifiedAccess =
    True(struct
	    let module_name = "UnspecifiedAccess"
	    let option_name = "-unspecified-access"
	    let descr = "assume that all read/write accesses occuring in unspecified order are not separated"
	  end)

  module ReadAnnot =
    True(struct
	   let module_name = "ReadAnnot"
	   let option_name = "-annot"
	   let descr = "read annotation"
	 end)
  let () =
    Clexer.annot_char := '@';
    ReadAnnot.add_set_hook
      (fun _ x ->
         (* prevent the C lexer interpretation of comments *)
	 if x then Clexer.annot_char := '@' else Clexer.annot_char := '\000')

  module PreprocessAnnot =
    False(struct
	    let module_name = "PreprocessAnnot"
	    let option_name = "-pp-annot"
	    let descr = "pre-process annotations (if they are read)"
	  end)

  module CppCommand =
    EmptyString
      (struct
	 let module_name = "CppCommand"
	 let option_name = "-cpp-command"
	 let arg_name = "cmd"
	 let descr = "<cmd> is used to build the preprocessing command.
	Defaults to $CPP environment variable or else \"gcc -C -E -I.\"
	If unset, the command is built as follow:
	CPP -o <preprocessed file> <source file>
	%1 and %2 can be used into CPP string to mark the position of <source file> and <preprocessed file> respectively"
       end)

  module CppExtraArgs =
    StringSet(struct
		let module_name = "CppExtraArgs"
		let option_name = "-cpp-extra-args"
		let arg_name = "args"
		let descr = "additional arguments passed to the preprocessor while preprocessing the C code but not while preprocessing annotations"
	      end)

  let () = Plugin.set_negative_option_name ""
  module TypeCheck =
    False(struct
            let module_name = "TypeCheck"
            let option_name = "-typecheck"
            let descr = "only typechecks the source files"
          end)

  module ContinueOnAnnotError =
    False(struct
            let module_name = "ContinueOnAnnotError"
            let option_name = "-continue-annot-error"
            let descr = "When an annotation fails to type-check, just emits \
                         a warning and discards the annotation instead of \
                         generating an error (errors in C are still fatal)"
          end)
  let () =
    ContinueOnAnnotError.add_set_hook
      (fun _ b ->
         if b then
           Cabshelper.continue_annot_error_set ()
         else Cabshelper.continue_annot_error_unset())

  (** {3 Journalization] *)
  module Journal = struct
    let () = set_negative_option_name "-journal-disable"
    module Enable =
      Bool
	(struct
	   let module_name = "Journal.Enable"
	   let default = Cmdline.journal_enable
	   let option_name = "-journal-enable"
	   let descr = "dump a journal while Frama-C exit"
	 end)
    module Name =
      String
	(struct
	   let module_name = "Journal.Name"
	   let option_name = "-journal-name"
	   let default = Journal.get_name ()
	   let arg_name = "s"
	   let descr =
	     "set the filename of the journal (do not write any extension)"
	 end)
  end

  let () = set_negative_option_name ""
  module NoType =
    Bool
      (struct
	 let module_name = "NoType"
	 let default = not Cmdline.use_type
	 let option_name = "-no-type"
	 let descr = "undocumented but disable some features"
       end)

  let () = set_negative_option_name ""
  module NoObj =
    Bool
      (struct
	 let module_name = "NoObj"
	 let default = not Cmdline.use_obj
	 let option_name = "-no-obj"
	 let descr = "-no-type + disable some additional features"
       end)


module SimplifyCfg =
  False(struct
	  let module_name = "SimplifyCfg"
	  let option_name = "-simplify-cfg"
	  let descr="remove break, continue and switch statement before analyzes"
	end)

module KeepSwitch =
  False(struct
	  let option_name = "-keep-switch"
	  let module_name = "KeepSwitch"
	  let descr = "keep switch statements despite -simplify-cfg"
	end)

module Constfold =
  False
    (struct
       let option_name = "-constfold"
       let module_name = "Constfold"
       let descr = "fold all constant expressions in the code before analysis"
     end)

module FloatDigits =
  Int(struct
	let module_name = "FloatDigits"
	let option_name = "-float-digits"
	let default = 4
	let arg_name = "n"
	let descr = "display this number of digits when printing floats. Defaults to 4"
      end)

module Files = struct
  include StringList
    (struct
       let option_name = ""
       let module_name = "Files"
       let arg_name = ""
       let descr = ""
     end)
  let () = Cmdline.use_cmdline_files set
  module Check =
    False(struct
	    let option_name = "-check"
	    let module_name = "Files.Check"
	    let descr = "performs consistency checks over cil files"
	  end)
  module Copy =
    False(struct
	    let option_name = "-copy"
	    let module_name = "Files.Copy"
	    let descr = "always perform a copy of the original AST before analysis begin"
	  end)
  module Orig_name =
    False(struct
	    let option_name = "-orig-name"
	    let module_name = "Files.Orig_name"
	    let descr = "prints a message each time a variable is renamed"
	  end)
end

(** {3 Viewer options} *)

(* [Pascal] For the reproducibility of bug reports, default values for options
   should not come from environment variables (that users will systematically
   forget to mention in their bug reports).
   The fonts are, with the installation path FRAMAC_SHARE, the only justifiable
   exception to this principle. *)
module MonospaceFontName =
  String(struct
	   let module_name = "MonospaceFontName"
	   let default =
	     try Sys.getenv "FRAMAC_MONOSPACEFONT"
	     with Not_found -> "Monospace,Lucida Sans Unicode,Sans"
	   let arg_name = "f"
	   let option_name = "-monospace-font"
	   let descr = "use font <f> as monospace font (defaults to the environment variable $FRAMAC_MONOSPACEFONT if defined and to Monospace,Lucida Sans Unicode,Sans otherwise)"
	 end)

module GeneralFontName =
  String(struct
	   let module_name = "GeneralFontName"
	   let default =
	     try Sys.getenv "FRAMAC_GENERALFONT"
	     with Not_found -> "Helvetica,Lucida Sans Unicode,Sans"
	   let option_name = "-general-font"
	   let arg_name = "f"
	   let descr = "use font f as general font (defaults to the environment variabe $FRAMAC_GENERALFONT if defined and to Helvetica,Lucida Sans Unicode,Sans otherwise)"
	 end)

end

(* ************************************************************************* *)
(** {2 Value analysis} *)
(* ************************************************************************* *)

(** General parameters *)
module Value =
  Plugin.Register
    (struct
       let name = "value analysis"
       let shortname = "value"
       let module_name = "Value"
       let descr =
	 "automatically computes variation domains for the variables of the program"
     end)






module ArrayPrecisionLevel =
  Value.Int
    (struct
       let default = 200
       let option_name = "-plevel"
       let arg_name = "n"
       let descr = "use <n> as the precision level for arrays accesses. Array accesses are precise as long as the interval for the index contains less than n values. (defaults to 200)"
     end)

module WideningLevel =
  Value.Int
    (struct
       let default = 3
       let option_name = "-wlevel"
       let arg_name = "n"
       let descr =
	 "set <n> as number of iterations before widening (defaults to 3)"
     end)

module AbsoluteValidRange = struct
  include Value.String
    (struct
       let option_name = "-absolute-valid-range"
       let arg_name = "min-max"
       let descr = "min and max must be integers in decimal, hexadecimal (0x, 0X), octal (0o) or binary (0b) notation and fit in 64 bits. Assume that that all absolute addresses outside of the [min-max] range are invalid. In the absence of this option, all absolute addresses are assumed to be invalid"
       let default = ""
     end)
  let is_set _x = assert false
end

module ForceMemzones =
  Value.False
    (struct
       let option_name = "-memzones"
       let descr = "force memzones display (undocumented)"
     end)

(* ************************************************************************* *)
(** {2 Interface for dynamic plugins} *)
(* ************************************************************************* *)

module Dynamic = struct

    module type Common = sig
      type t
      val get: string -> t
      val set: string -> t  -> unit
      val clear: string -> unit -> unit
      val is_set: string  -> bool
      val is_default: string  -> bool
    end

    let apply modname name s ty1 ty2 =
      let name = dynamic_funname modname s name in
      Dynamic.get name (Type.func ty1 ty2)

    module Common(X: sig type t val modname:string val ty: t Type.t end ) =
    struct
      type t = X.t
      let ty = X.ty
      let get name = apply X.modname name "get" Type.unit ty ()
      let set name = apply X.modname name "set" ty Type.unit
      let clear name = apply X.modname name "clear" Type.unit Type.unit
      let is_set name = apply X.modname name "is_set" Type.unit Type.bool ()
      let is_default name =
	apply X.modname name "is_default" Type.unit Type.bool ()
    end

    module Bool = struct
      include Common
	(struct type t = bool let ty = Type.bool let modname = "Bool"end )
      let on name = apply "Bool" name "on" Type.unit Type.unit
      let off name = apply "Bool" name "off" Type.unit Type.unit
    end

    module Int = struct
      include Common
	(struct type t = int let ty = Type.int let modname = "Int" end )
      let incr name = apply "Int" name "incr" Type.unit Type.unit
    end

    module String =
      Common
	(struct
	   type t = string
	   let ty = Type.string
	   let modname = "String"
	 end)

    module StringSet = struct
      include Common
	(struct
	   type t = Cilutil.StringSet.t
	   let ty = Kernel_type.string_set
	   let modname = "StringSet"
	 end)
      let add name = apply "StringSet" name "add" Type.string Type.unit
      let add_set name = apply "StringSet" name "add_set" Type.string Type.unit
      let is_empty name =
	apply "StringSet" name "is_empty" Type.unit Type.bool ()
      let iter name =
	apply "StringSet" name "iter"
	  (Type.func Type.string Type.unit) Type.unit
(*      let fold _name = assert false*)
    end
(*
    module IndexedVal(X: sig val ty_name: string end) = struct
      include Common(struct type t = string let ty = string end)
      type value = Type.ty
      let ty = Type.get_abstract X.ty_name
      let add_choice name =
	StringTbl.find tbl (name ^ ".add_choice") (func string (func ty unit))
      let get_val name =
	StringTbl.find tbl (name ^ ".get_val") (func unit ty) ()
    end
      *)

end

(* ************************************************************************* *)

include General
    (* Do this at end in order to not add misleading stuff in the environment
       before *)

(** {2 Options which define context of analyses } *)

let get_selection_context () =
  let a o = Project.Selection.add o Kind.Do_Not_Select_Dependencies in
  let has_dependencies o =
    try
      Project.Selection.iter
	(fun _ -> raise Exit)
	(Project.Selection.singleton o Kind.Only_Select_Dependencies);
      false
    with Exit ->
      true
  in
  (* automatically select all options which have some dependencies:
     they have an impact of some analysis. *)
  let sel_ctx =
    Project.Selection.fold
      (fun o _ acc -> if has_dependencies o then a o acc else acc)
      (get_selection ())
      Project.Selection.empty
  in
  (* Value analysis *)
  let sel_ctx = a WideningLevel.self sel_ctx in
  let sel_ctx = a FloatDigits.self sel_ctx in
  (* General options *)
  ((a Files.Check.self) $
   (a Files.Copy.self))
    sel_ctx

(*
Local Variables:
compile-command: "LC_ALL=C make -C ../.."
End:
*)
